# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
import contextlib
from copy import copy
from itertools import chain
from inspect import getfullargspec as getargspec

from .nf import GeneratorLimitExceeded, symmGB_F2_C, symmGB_F2_python
from .pbori import GroebnerStrategy, ll_red_nf_redsb
from .PyPolyBoRi import (Monomial, Polynomial,
                         OrderCode)
from .ll import eliminate, ll_encode
from .statistics import used_vars_set
from .heuristics import dense_system, gauss_on_linear
from .easy_polynomials import easy_linear_polynomials
from .interpolate import lex_groebner_basis_for_polynomial_via_variety
from .fglm import _fglm


def get_options_from_function(f):
    (argnames, varargs, varopts, defaults) = getargspec(f)[:4]
    return dict(zip(argnames[-len(defaults):], defaults))


def filter_oldstyle_options(**options):
    filtered = {}
    for key in options:
        newkey = key
        for prefix in ['use_', 'opt_allow_', 'opt_']:
            newkey = newkey.replace(prefix, '')
        filtered[newkey] = options[key]
    return filtered


def filter_newstyle_options(func, **options):
    allowed = get_options_from_function(func).keys()
    filtered = {}
    for key in options:
        for prefix in ['', 'use_', 'opt_', 'opt_allow_']:
            if prefix + key in allowed:
                filtered[prefix + key] = options[key]

    return filtered


def want_interpolation_gb(G):
    if not G or G[0].ring().get_order_code() != OrderCode.lp or len(G) != 1:
        return False
    p = Polynomial(G[0])
    return not (p.lead_deg() <= 1 or p.set().n_nodes() > 1000)


def ll_is_good(I):
    lex_lead = set()
    for p in I:
        if not p.is_zero():
            m = p.lex_lead()
            if m.deg() == 1:
                lex_lead.add(next(iter(m.variables())).index())
    if len(lex_lead) >= 0.8 * len(I):
        uv = used_vars_set(I).deg()  # don't use len here, which will yield 1
        if len(lex_lead) > 0.9 * uv:
            if uv - len(lex_lead) > 16:
                return "llfirstonthefly"
            return "llfirst"
    return False


def ll_heuristic(d):
    d = copy(d)
    I = d["I"]
    if ("llfirstonthefly" not in d) and ("llfirst" not in d):
        hint = ll_is_good(I)
        if hint:
            d[hint] = True
    return d


def change_order_heuristic(d):
    d_orig = d
    d = copy(d)
    I = d["I"]
    if not I:
        return d
    switch_table = {OrderCode.lp: OrderCode.dp_asc, OrderCode.dlex: OrderCode.
                    dp_asc}
    if "other_ordering_first" not in d:
        # TODO after ll situation might look much different, so heuristic is on
        # wrong place
        code = next(iter(I)).ring().get_order_code()
        if code in switch_table:
            max_non_linear = len(I) // 2
            non_linear = 0
            if code == OrderCode.lp:
                for p in I:
                    if p.lead_deg() > 1:
                        non_linear = non_linear + 1
                        if non_linear > max_non_linear:
                            break
            if (non_linear > max_non_linear) or (code != OrderCode.lp):
                other_ordering_opts = copy(d_orig)
                other_ordering_opts["switch_to"] = switch_table[code]
                d["other_ordering_first"] = other_ordering_opts
    return d


def interpolation_gb_heuristic(d):
    d = copy(d)
    I = d["I"]
    if not d.get("other_ordering_opts", False) and want_interpolation_gb(I):
        d["interpolation_gb"] = True
        d["other_ordering_first"] = False
    return d


def linear_algebra_heuristic(d):
    d = copy(d)
    I = d["I"]

    def want_la():
        if not I:
            return False
        n_used_vars = None
        bound = None
        if next(iter(I)).ring().has_degree_order():
            new_bound = 200
            n_used_vars = used_vars_set(I, bound=new_bound).deg()
            if n_used_vars < new_bound:
                return True
            bound = new_bound
        if dense_system(I):
            new_bound = 100
            if not (bound and new_bound < bound):
                n_used_vars = used_vars_set(I, bound=new_bound).deg()
                bound = new_bound
            if n_used_vars < bound:
                return True
        return False
    if not (("faugere" in d and not d["faugere"]) or
            ("noro" in d and d["noro"])):
        if ("faugere" in d and d["faugere"]) or want_la():

            d["faugere"] = True
            if "red_tail" not in d:
                d["red_tail"] = False
            if "selection_size" not in d:
                d["selection_size"] = 10000
            if "ll" not in d:
                d["ll"] = True

    return d


def trivial_heuristic(d):
    return d


class HeuristicalFunction:
    def __call__(self, *args, **kwds):
        complete_dict = copy(kwds)
        heuristic = True
        with contextlib.suppress(KeyError):
            heuristic = complete_dict["heuristic"]
        for (k, v) in zip(self.argnames, args):
            complete_dict[k] = v
        if heuristic:
            complete_dict = self.heuristicFunction(complete_dict)
        return self.f(**complete_dict)

    def __init__(self, f, heuristic_function):
        self.argnames, self.varargs, self.varopts, self.defaults = getargspec(f)[:4]
        if hasattr(f, "options"):
            self.options = f.options
        else:
            self.options = dict(zip(self.argnames[-len(self.defaults):], self.
                                    defaults))
        self.heuristicFunction = heuristic_function
        self.f = f
        self.__doc__ = f.__doc__


def with_heuristic(heuristic_function):
    def make_wrapper(f):
        wrapped = HeuristicalFunction(f, heuristic_function)
        wrapped.__name__ = f.__name__
        return wrapped
    return make_wrapper


def clean_polys_pre(I):
    wrap = (Polynomial(p) for p in I)
    return (list({p for p in wrap if not p.is_zero()}), None)


def gb_with_pre_post_option(option, pre=None,
                            post=None, if_not_option=(),
                            default=False):
    def make_wrapper(f):
        def wrapper(I, **kwds):
            prot = kwds.get("prot", False)
            for o in if_not_option:
                if (o in kwds and kwds[o]) or (o not in kwds and
                                               groebner_basis.options[o]):
                    option_set = False
            if "option_set" not in locals():
                option_set = kwds.get(option, default)
            kwds = {o: kwds[o] for o in kwds if o != option}
            state = None

            if option_set and pre:
                pre_args = getargspec(pre)[0]
                if prot:
                    print("preprocessing for option:", option)

                local_symbols = copy(locals())
                (I, state) = pre(**{k: v for (k, v) in local_symbols.items()
                                    if k in pre_args})
            I = f(I, **kwds)
            if option_set and post:
                post_args = getargspec(post)[0]
                if prot:
                    print("postprocessing for option:", option)
                local_symbols = copy(locals())
                I = post(**{k: v for (k, v) in local_symbols.items()
                            if k in post_args})

            return I
        wrapper.__name__ = f.__name__
        wrapper.__doc__ = f.__doc__
        if hasattr(f, "options"):
            wrapper.options = copy(f.options)
        else:

            wrapper.options = get_options_from_function(f)

        wrapper.options[option] = default
        return wrapper
    return make_wrapper


def redsb_post(I, state):
    if not I:
        return []
    return I.minimalize_and_tail_reduce()


def minsb_post(I, state):
    if not I:
        return []
    return I.minimalize()


def invert_all(I):
    return [p.map_every_x_to_x_plus_one() for p in I]


def invert_all_pre(I):
    return (invert_all(I), None)


def invert_all_post(I, state):
    return invert_all(I)


def llfirst_pre(I, prot):
    (eliminated, llnf, I) = eliminate(I, on_the_fly=False, prot=prot)
    return (I, eliminated)


def ll_constants_pre(I):
    ll_res = []

    while any(p.lex_lead_deg() == 1 and (p + p.lex_lead()).constant()
              for p in I):
        I_new = []
        ll = []
        leads = set()
        for p in I:
            if p.lex_lead_deg() == 1:
                l = p.lead()
                if l not in leads and p.is_singleton_or_pair():
                    tail = p + l
                    if tail.deg() <= 0:
                        ll.append(p)
                        leads.add(l)
                        continue
            I_new.append(p)
        encoded = ll_encode(ll)
        reduced = []
        for p in I_new:
            rp = ll_red_nf_redsb(p, encoded)
            if not rp.is_zero():
                reduced.append(rp)
        I = reduced
        ll_res.extend(ll)
    return (I, ll_res)


def variety_size_from_gb(I):
    """
    TESTS::

        sage: from sage.rings.polynomial.pbori import Ring, Monomial, Polynomial
        sage: from sage.rings.polynomial.pbori.gbcore import variety_size_from_gb
        sage: r = Ring(100)
        sage: x = r.variable
        sage: variety_size_from_gb([])
        1
        sage: variety_size_from_gb([Polynomial(0, r)])
        1
        sage: variety_size_from_gb([Polynomial(1, r)])
        0.0
        sage: variety_size_from_gb([x(1)])
        1.0
        sage: variety_size_from_gb([x(1), x(2)])
        1.0
        sage: variety_size_from_gb([x(1), x(2)*x(3)])
        3.0
        sage: variety_size_from_gb([x(1), x(1)*x(4), x(2)*x(3)])
        6.0
        sage: variety_size_from_gb([x(1)*x(2), x(2)*x(3)])
        5.0
        sage: mons = [Monomial([r.variable(i) for i in range(100) if i!=j])\
        ....: for j in range(100)]
        sage: variety_size_from_gb(mons)
        1.2676506002282294e+30
    """
    I = (Polynomial(p) for p in I)
    I = [p for p in I if not p.is_zero()]
    if not I:
        return 1
#     # TODO Here's something wrong! See the example with 5 solutions.
#     # (reverting for now)
#     number_of_used_vars = used_vars_set(I).deg()
#     leads = set([p.lead() for p in I])
#     minimal_leads = BooleSet(leads).minimal_elements()
#     number_of_used_vars_minimal_leads =\
#         minimal_leads.vars().deg()
#     standard_monomials =\
#         minimal_leads.include_divisors().diff(minimal_leads)
#     return standard_monomials.size_double()*\
#         2**(number_of_used_vars-number_of_used_vars_minimal_leads)

    sm = Monomial(used_vars_set(I)).divisors()
    for p in I:
        m = p.lead()
        sm = sm.diff(sm.multiples_of(m))
    return sm.size_double()


def other_ordering_pre(I, option_set, kwds):
    """
    TESTS::

        sage: from sage.rings.polynomial.pbori.blocks import declare_ring
        sage: r = declare_ring(['x0', 'x1', 'x2', 'x3', 'x4'], globals())
        sage: id = [x1*x3 + x1 + x2*x3 + x3 + x4, x0*x3 + x0 + x1*x2 + x2 + 1,  x1*x3 + x1*x4 + x3*x4 + x4 + 1, x0*x2 + x0*x4 + x1 + x3 + x4]
        sage: from sage.rings.polynomial.pbori.gbcore import groebner_basis
        sage: groebner_basis(id)
        [1]
    """
    if not I:
        return (I, None)

    main_kwds = kwds
    options = option_set

    old_ring = next(iter(I)).ring()
    try:
        new_ring = old_ring.clone(ordering=options["switch_to"])

        kwds = {k: options[k] for k in options
                if k not in ("other_ordering_first", "switch_to", "I")}
        kwds["redsb"] = True
        I = groebner_basis([new_ring(poly) for poly in I], **kwds)
        variety_size = variety_size_from_gb(I)

        fglm_bound = options.get("fglm_bound") or groebner_basis.options["fglm_bound"]
        if variety_size < fglm_bound:
            main_kwds["convert_with_fglm_from_ring"] = new_ring
            main_kwds["convert_with_fglm_to_ring"] = old_ring
        else:
            I = [old_ring(poly) for poly in I]
    finally:
        pass

    return (I, None)


def llfirstonthefly_pre(I, prot):
    (eliminated, llnf, I) = eliminate(I, on_the_fly=True)
    return (I, eliminated)


def gauss_on_linear_pre(I, prot):
    return (gauss_on_linear(I), None)


def easy_linear_polynomials_pre(I):
    res = []
    for p in I:
        res.append(p)
        res.extend(easy_linear_polynomials(p))

    return (list(set(res)), None)


def llfirst_post(I, state, prot, kwds):
    eliminated = state
    for p in I:
        if p.is_one():
            return [p]

    if eliminated:
        I = list(chain(I, eliminated))
        # redsb just for safety, as don't know how option is set
        kwds = copy(kwds)
        kwds.update(
            {'llfirst': False,
             'llfirstonthefly': False,
             'll_constants': False,
             'deg_bound': False,
             'other_ordering_first': False,
             'eliminate_identical_variables': False, 'redsb': True})
        I = groebner_basis(I, **kwds)
    return I


def ll_constants_post(I, state):
    eliminated = state
    for p in I:
        if p.is_one():
            return [p]
    if eliminated:
        I = list(chain(I, eliminated))
        # redsb just for safety, as don't know how option is set
    return I


def result_to_list_post(I, state):
    return list(I)


def fix_deg_bound_post(I, state):
    if isinstance(I, GroebnerStrategy):
        return I.all_generators()
    return I


def incremental_pre(I, prot, kwds):
    def sort_key(p):
        p = Polynomial(p)
        return (p.navigation().value(), -p.deg())
    I = sorted(I, key=sort_key)
    inc_sys = []
    kwds = copy(kwds)
    kwds['incremental'] = False

    for p in I[:-1]:
        inc_sys.append(p)
        inc_sys = groebner_basis(inc_sys, **kwds)
        if prot:
            print("incrementally calculating GB, adding generator:", p)
    inc_sys.append(I[:-1])
    return (inc_sys, None)


def eliminate_identical_variables_pre(I, prot):
    changed = True
    ll_system = []
    treated_linears = set()
    while changed:
        changed = False
        rules = {}
        for p in I:
            t = p + p.lead()
            if p.lead_deg() == 1:
                l = p.lead()
                if l in treated_linears:
                    continue
                treated_linears.add(l)
                if t.deg() > 0:
                    rules.setdefault(t, [])
                    leads = rules[t]
                    leads.append(l)

        def my_sort_key(l):
            return l.navigation().value()

        for t, leads in rules.items():
            if len(leads) > 1:
                changed = True
                sleads = sorted(leads, key=my_sort_key, reverse=True)
                chosen = sleads[0]
                ll_system.extend(chosen + v for v in sleads[1:])
    if ll_system:
        ll_encoded = ll_encode(ll_system, reduce=True)
        I = {ll_red_nf_redsb(p, ll_encoded) for p in I}
    return (I, ll_system)


@gb_with_pre_post_option("clean_arguments", pre=clean_polys_pre, default=True)
@gb_with_pre_post_option("easy_linear_polynomials",
                         pre=easy_linear_polynomials_pre, default=True)
@gb_with_pre_post_option("result_to_list", post=result_to_list_post,
                         default=True)
@with_heuristic(interpolation_gb_heuristic)
@gb_with_pre_post_option("invert", pre=invert_all_pre,
                         post=invert_all_post, default=False)
@gb_with_pre_post_option("gauss_on_linear", pre=gauss_on_linear_pre,
                         default=True)
@gb_with_pre_post_option("ll_constants", pre=ll_constants_pre,
                         post=ll_constants_post, default=True)
@gb_with_pre_post_option("eliminate_identical_variables",
                         pre=eliminate_identical_variables_pre,
                         post=llfirst_post, default=True)
@with_heuristic(ll_heuristic)
@gb_with_pre_post_option("llfirst", if_not_option=["llfirstonthefly"],
                         pre=llfirst_pre, post=llfirst_post, default=False)
@gb_with_pre_post_option("llfirstonthefly", pre=llfirstonthefly_pre,
                         post=llfirst_post, default=False)
@gb_with_pre_post_option("incremental", pre=incremental_pre)
@with_heuristic(change_order_heuristic)
@gb_with_pre_post_option("other_ordering_first", if_not_option=[
    "interpolation_gb"], pre=other_ordering_pre, default=False)
@with_heuristic(linear_algebra_heuristic)
@gb_with_pre_post_option("fix_deg_bound", if_not_option=["interpolation_gb"],
                         post=fix_deg_bound_post, default=True)
@gb_with_pre_post_option("minsb", post=minsb_post,
                         if_not_option=["redsb", "deg_bound",
                                        "interpolation_gb",
                                        "convert_with_fglm_from_ring"],
                         default=True)
@gb_with_pre_post_option("redsb", post=redsb_post,
                         if_not_option=["deg_bound", "interpolation_gb",
                                        "convert_with_fglm_from_ring"],
                         default=True)
def groebner_basis(I, heuristic=True, unique_ideal_generator=False,
                   interpolation_gb=False, clean_and_restart_algorithm=False,
                   convert_with_fglm_from_ring=None,
                   convert_with_fglm_to_ring=None,
                   fglm_bound=40000,
                   modified_linear_algebra=True, preprocessor=None,
                   deg_bound=False,
                   implementation='Python', full_prot=False, prot=False,
                   draw_matrices=False, preprocess_only=False, **impl_options):
    """Computes a Groebner basis of a given ideal I, w.r.t options."""

    if not I:
        return I

    if full_prot:
        prot = True
    if prot:
        print("number of passed generators:", len(I))
    if convert_with_fglm_from_ring is not None:
        from_ring = convert_with_fglm_from_ring
        to_ring = convert_with_fglm_to_ring
        return _fglm(I, from_ring, to_ring)

    if interpolation_gb:
        first = next(iter(I))
        if len(I) != 1 or first.ring().get_order_code() != OrderCode.lp:
            raise ValueError
        return lex_groebner_basis_for_polynomial_via_variety(first)
    if deg_bound is False:
        deg_bound = 100000000
    I = [Polynomial(p) for p in I if not p.is_zero()]
    if unique_ideal_generator and I:
        prod = 1
        for p in I:
            prod = (p + 1) * prod
        I = [prod + 1]

    implementation = symmGB_F2_python if implementation == 'Python' else symmGB_F2_C

    # custom preprocessing
    if preprocessor:
        I = preprocessor(I)

    if preprocess_only:
        for p in I:
            print(p)
        import sys
        sys.exit(0)

    def call_algorithm(I, max_generators=None):
        return implementation(I,
                              deg_bound=deg_bound,
                              full_prot=full_prot,
                              prot=prot,
                              max_generators=max_generators,
                              draw_matrices=draw_matrices,
                              **filter_newstyle_options(implementation,
                                                        **impl_options))

    if clean_and_restart_algorithm:
        for max_generators in [1000, 10000, 50000, 100000, 200000, 300000,
                               400000, None]:
            try:
                return call_algorithm(I, max_generators=max_generators)
            except GeneratorLimitExceeded as e:
                I = list(e.strat.all_generators())
                del e.strat
                if prot:
                    print("generator limit exceeded:", max_generators,
                          "restarting algorithm")
    else:
        return call_algorithm(I)


def build_groebner_basis_doc_string():
    additional_options_from_buchberger = filter_oldstyle_options(
        **get_options_from_function(symmGB_F2_python))
    for k in list(additional_options_from_buchberger):
        if k in groebner_basis.options:
            del additional_options_from_buchberger[k]

    gdoc = groebner_basis.__doc__
    gdoc += "\nOptions are:\n"
    gdoc += "\n".join(k + "  :  " + repr(groebner_basis.options[k])
                      for k in groebner_basis.options)
    gdoc += """

Turn off heuristic by setting heuristic=False
  Additional options come from the actual buchberger implementation.
  In case of our standard Python implementation these are the following:
"""
    gdoc += "\n".join(k + "  :  " + repr(additional_options_from_buchberger[k])
                      for k in additional_options_from_buchberger)
    groebner_basis.__doc__ = gdoc


build_groebner_basis_doc_string()



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from .pbori import ReductionStrategy
from .PyPolyBoRi import Polynomial


def interred(l, completely=False):
    r"""
    Compute a new generating system (g1, ...,gn),
    spanning the same ideal modulo field equations.

    The system is interreduced: For i!=j:
    gi.lead() does not divide any leading term of gj.

    If completely is set to ``True``, then also terms in the
    tail are not reducible by other polynomials.
    """
    l = [Polynomial(p) for p in l if p != 0]
    if not l:
        return []
    ring = l[0].ring()
    l_old = None
    l = tuple(l)
    while l_old != l:
        l_old = l
        l = sorted(l, key=Polynomial.lead)
        g = ReductionStrategy(ring)
        if completely:
            g.opt_red_tail = True
        for p in l:
            gp = g.nf(p)
            if not gp.is_zero():
                g.add_generator(gp)
        l = tuple(e.p for e in g)
    return l



# sage_setup: distribution = sagemath-brial
# distutils: language = c++
# distutils: extra_compile_args = -std=c++17

from libcpp.string cimport string as std_string
from libcpp.vector cimport vector


cdef extern from "polybori/pbori_defs.h" namespace "COrderEnums":
    cdef enum ordercodes:
        pblp            "CTypes::lp"
        pbdlex          "CTypes::dlex"
        pbdp_asc        "CTypes::dp_asc"
        pbblock_dlex    "CTypes::block_dlex"
        pbblock_dp_asc  "CTypes::block_dp_asc"
        pbdp            "static_cast<COrderEnums::ordercodes>(17)"
        pbblock_dp      "static_cast<COrderEnums::ordercodes>(19)"

    cdef enum comparecodes:
        less_than               "CTypes::less_than"
        equality                "CTypes::equality"
        greater_than            "CTypes::greater_than"
        less_or_equal_max       "CTypes::less_or_equal_max"
        greater_or_equal_min    "CTypes::greater_or_equal_min"


cdef extern from "pb_wrap.h":
    cdef cppclass PBNavigator "CCuddNavigator":
        PBNavigator thenBranch()
        PBNavigator elseBranch()
        int value "operator*"()
        bint isConstant()
        bint isTerminated()
        bint operator==(const PBNavigator&)
        size_t hash()

    cdef cppclass PBBlockIter "COrderingBase::block_iterator":
        long dereference "operator*"()
        void increment "operator++"()
        bint operator!=(const PBBlockIter&)

    cdef cppclass PBOrdering "COrderingBase":
        int getOrderCode()
        int getBaseOrderCode()
        void appendBlock(int nextBlockStartIndex)
        PBBlockIter blockBegin()
        PBBlockIter blockEnd()
        int lastBlockStart()
        bint isDegreeOrder()

    cdef cppclass PBSet "DefaultRinged<BooleSet> "
    cdef cppclass PBPoly "DefaultRinged<BoolePolynomial> "

    cdef cppclass PBBooleVariable "BooleVariable":
        PBBooleVariable(int idx, const PBRing& r)

    cdef cppclass PBVar "DefaultRinged<BooleVariable> ":
        void operator=(const PBBooleVariable&)
        int index()
        bint operator==(const PBVar&)

    cdef cppclass PBRing "BoolePolyRing":
        PBRing()
        PBRing(int nvars, int order)
        size_t id()
        size_t nVariables()
        PBVar variable(int n)
        PBOrdering ordering()
        void changeOrdering(int)
        size_t hash()
        void setVariableName(int idx, char *varname)
        char* getVariableName(int idx)
        PBRing clone()
        PBPoly coerce(PBPoly rhs)
        PBPoly zero()
        PBPoly one()

    cdef cppclass PBVarBlock "VariableBlock":
        PBVarBlock(int size, int start_index, int offset, bint reverse,
                   PBRing r)
        PBVar operator()(int index)

    cdef cppclass PBMonomIter "BooleMonomial::const_iterator":
        int dereference()
        void increment()
        bint operator==(const PBMonomIter&)

    cdef cppclass PBMonomVarIter "BooleMonomial::variable_iterator":
        PBVar dereference()
        void increment()
        bint operator==(const PBMonomVarIter&)

    cdef cppclass PBMonom "DefaultRinged<BooleMonomial> ":
        PBMonom()
        PBMonom(PBRing r)
        PBMonom(PBVar m)
        void operator=(const PBBooleSet&)
        bint reducibleBy(PBMonom rhs)
        int deg()
        size_t hash()
        size_t stableHash()
        int firstIndex()
        comparecodes compare(PBMonom rhs)
        PBSet set()
        PBSet divisors()
        PBSet multiples(PBMonom rhs)
        PBMonomIter begin()
        PBMonomIter end()
        PBMonomVarIter variableBegin()
        PBMonomVarIter variableEnd()
        void imul "operator*="(PBMonom right)
        void idiv "operator/="(PBMonom right)
        PBNavigator navigation()
        PBMonom GCD(PBMonom rhs)
        PBRing ring()

    cdef cppclass PBSetIter "BooleSet::const_iterator":
        PBSetIter(const PBSetIter&)
        PBMonom dereference()
        void increment()
        bint operator==(const PBSetIter&)

    cdef cppclass PBBooleSet "BooleSet":
        PBBooleSet()
        PBBooleSet(PBRing r)
        PBBooleSet(PBPoly p)
        PBBooleSet(PBNavigator d, PBRing r)
        PBBooleSet(int ind, PBNavigator a, PBNavigator b, PBRing r)

    cdef cppclass PBSet "DefaultRinged<BooleSet> ":
        void operator=(const PBBooleSet&)
        bint owns(PBMonom val)
        int nNodes()
        int nSupport()
        int size()
        size_t hash()
        size_t stableHash()
        PBNavigator navigation()
        PBSet cartesianProduct(PBSet rhs)
        PBSet diff(PBSet rhs)
        PBSet divide(PBMonom rhs)
        PBSet change(int idx)
        PBSet subset0(int i)
        PBSet subset1(int i)
        PBSet unite(PBSet rhs)
        PBSet intersect(PBSet rhs)
        PBMonom usedVariables()
        PBSet divisorsOf(PBMonom rhs)
        PBSet multiplesOf(PBMonom rhs)
        double sizeDouble()
        PBSetIter begin()
        PBSetIter end()
        bint isZero()
        bint isOne()

    PBSet pb_include_divisors "include_divisors"(PBSet p)
    PBSet pb_minimal_elements "minimal_elements"(PBSet p)

    cdef cppclass PBPolyIter "BoolePolynomial::ordered_iterator":
        PBPolyIter(const PBPolyIter&)
        PBMonom dereference()
        void increment()
        bint operator==(const PBPolyIter&)

    cdef cppclass PBBoolePolynomial "BoolePolynomial":
        void operator=(const PBPoly&)
        PBBoolePolynomial(PBRing r)
        PBBoolePolynomial(PBSet d)
        PBBoolePolynomial(PBMonom d)
        PBBoolePolynomial(PBVar d)
        PBBoolePolynomial(int d, PBRing r)

    cdef cppclass PBPoly "DefaultRinged<BoolePolynomial> ":
        PBPoly()
        void operator=(const PBBoolePolynomial&)
        int deg()
        int leadDeg()
        int lexLeadDeg()
        int totalDeg()
        int length()
        int eliminationLength()
        int nUsedVariables()
        int nNodes()
        size_t hash()
        size_t stableHash()
        bint isZero()
        bint isOne()
        bint isConstant()
        bint isSingleton()
        bint isPair()
        bint isSingletonOrPair()
        bint hasConstantPart()
        bint firstReducibleBy(PBPoly rhs)
        PBMonom lead()
        PBMonom lexLead()
        PBMonom firstTerm()
        PBMonom usedVariables()
        PBPoly gradedPart(int d)
        PBSet diagram()
        PBSet set()
        PBSet leadDivisors()
        PBNavigator navigation()
        PBPolyIter orderedBegin()
        PBPolyIter orderedEnd()
        void iadd "operator+="(PBPoly right)
        void iadd_PBMonom "operator+="(PBMonom right)
        void imul "operator*="(PBPoly right)
        void imul_monom "operator*="(PBMonom right)
        void idiv "operator/="(PBPoly right)
        void idiv_monom "operator/="(PBMonom right)
        bint operator==(const PBPoly&)

    PBSet pb_zeros "zeros"(PBPoly p, PBSet s)
    PBPoly pb_spoly "spoly"(PBPoly p, PBPoly r)
    PBPoly pb_map_every_x_to_x_plus_one "map_every_x_to_x_plus_one"(PBPoly)


ctypedef vector[PBBoolePolynomial] PBPolyVector
ctypedef vector[PBBoolePolynomial].iterator PBPolyVectorIter


cdef extern from *:
    cdef cppclass PBPolyEntry "PolyEntry":
        PBPoly p

    cdef cppclass PBRedStrategy "ReductionStrategy":
        PBRedStrategy()
        PBRedStrategy(const PBRing&)
        PBSet leadingTerms
        PBSet minimalLeadingTerms
        PBSet leadingTerms11
        PBSet leadingTerms00
        PBSet llReductor
        PBSet monomials
        PBSet monomials_plus_one
        bint optBrutalReductions
        bint optLL
        PBPoly nf(PBPoly p)
        bint optRedTailDegGrowth
        bint optRedTail
        void setupSetsForLastElement()
        bint canRewrite(PBPoly p)
        void addGenerator(PBPoly p)
        int select1(PBMonom p)

        int select_short(PBPoly p)
        PBPoly headNormalForm(PBPoly p)
        PBPoly reducedNormalForm(PBPoly p)

        size_t size()
        PBPolyEntry operator[](int)

    cdef cppclass PBFGLMStrategy "FGLMStrategy":
        PBFGLMStrategy(PBRing from_ring, PBRing to_ring, PBPolyVector vec)
        PBPolyVector main()

    cdef cppclass PBGBStrategy "GroebnerStrategy":
        PBGBStrategy(PBRing)
        bint reduceByTailReduced
        PBRedStrategy generators

        bint enabledLog
        unsigned int reductionSteps
        int normalForms
        int currentDegree
        int chainCriterions
        int variableChainCriterions
        int easyProductCriterions
        int extendedProductCriterions
        int averageLength

        bint optHFE
        bint optLazy

        bint optDelayNonMinimals

        bint optExchange
        bint optAllowRecursion

        bint optStepBounded
        bint optLinearAlgebraInLastBlock
        bint optModifiedLinearAlgebra
        bint optRedTailInLastBlock
        bint optDrawMatrices
        std_string matrixPrefix

        PBSet llReductor
        PBSet minimalLeadingTerms
        PBSet leadingTerms
        bint containsOne()
        bint variableHasValue(int idx)
        int addGenerator(PBPoly, bint is_impl)
        void addGeneratorDelayed(PBPoly)
        void addAsYouWish(PBPoly)
        void addNonTrivialImplicationsDelayed(const PBPolyEntry&)
        void symmGB_F2()
        void llReduceAll()
        void cleanTopByChainCriterion "pairs.cleanTopByChainCriterion"()
        size_t npairs "pairs.queue.size"()
        int suggestPluginVariable()
        PBPoly nextSpoly()
        PBPoly nf(PBPoly p)
        PBPolyVector allGenerators()
        PBPolyVector minimalize()
        PBPolyVector minimalizeAndTailReduce()
        PBPolyVector faugereStepDense(PBPolyVector v)

    int pairs_top_sugar(PBGBStrategy strat)
    PBPolyVector someNextDegreeSpolys(PBGBStrategy strat, size_t n)
    PBPolyVector nextDegreeSpolys(PBGBStrategy strat)
    PBPolyVector small_next_degree_spolys(PBGBStrategy strat, double f, size_t n)
    void implications(PBGBStrategy strat, int i)

    PBPoly cheap_reductions(PBRedStrategy strat, PBPoly p)

    PBPoly pb_add_up_polynomials "add_up_polynomials" \
        (PBPolyVector v, PBPoly p)

    PBPoly pb_nf3 "nf3" (PBRedStrategy strat, PBPoly p, PBMonom m)

    PBPoly pb_red_tail "red_tail" (PBRedStrategy strat, PBPoly p)

    PBPoly pb_interpolate "interpolate" (PBSet z, PBSet o)

    PBPoly pb_interpolate_smallest_lex "interpolate_smallest_lex" \
        (PBSet z, PBSet o)

    PBSet pb_contained_variables_cudd_style "contained_variables_cudd_style" \
        (PBSet m)

    PBSet pb_mod_var_set "mod_var_set" (PBSet a, PBSet v)

    PBPoly pb_mult_fast_sim "mult_fast_sim" (PBPolyVector v, PBRing r)

    PBSet pb_recursively_insert "recursively_insert"(PBNavigator p,
                                                int idx, PBSet m)
    PBPoly pb_ll_red_nf_noredsb "ll_red_nf_noredsb"(PBPoly p,
                                                PBSet reductors)
    PBPoly pb_ll_red_nf_noredsb_single_recursive_call "ll_red_nf_noredsb_single_recursive_call" \
        (PBPoly p, PBSet reductors)

    PBPoly pb_ll_red_nf "ll_red_nf"(PBPoly p, PBSet reductors)

    PBSet pb_mod_mon_set "mod_mon_set"(PBSet as, PBSet vs)

    PBPolyVector pb_parallel_reduce "parallel_reduce" \
        (PBPolyVector inp, PBGBStrategy strat, int average_steps, double delay_f)

    PBPolyVector pb_gauss_on_polys "gauss_on_polys" (PBPolyVector inp)

    PBPolyVector pb_easy_linear_factors "easy_linear_factors"(PBPoly p)

    PBPoly pb_substitute_variables "substitute_variables<BoolePolyRing, std::vector<BoolePolynomial>, BoolePolynomial>" \
        (PBRing ring, PBPolyVector vec, PBPoly poly) except +

    void pb_set_variable_name "BooleEnv::setVariableName" \
        (int idx, char *varname)

    char * pb_get_variable_name "BooleEnv::getVariableName" \
        (int idx)

    PBSet pb_random_set "random_set" (PBMonom variables, int length)
    void pb_set_random_seed "set_random_seed" (unsigned int seed)

    void PBPolyVector_set(PBPolyVector v, int i, PBPoly p)

    cdef cppclass PBConstant "BooleConstant":
        PBConstant()
        PBConstant(int val)
        bint isZero()
        bint isOne()
        bint isConstant()
        bint hasConstantPart()
        int deg()

    cdef cppclass PBVariableFactory "VariableFactory":
        PBVariableFactory(const PBRing&)

    cdef cppclass PBVarFactory "DefaultRinged<VariableFactory> ":
        void operator=(const PBVariableFactory&)
        PBVar operator()(long index)

    cdef cppclass PBMonomialFactory "MonomialFactory":
        PBMonomialFactory(const PBRing&)

    cdef cppclass PBMonomFactory "DefaultRinged<MonomialFactory> ":
        void operator=(const PBMonomialFactory&)
        PBMonom operator()()

    cdef cppclass PBPolynomialFactory "PolynomialFactory":
        PBPolynomialFactory(const PBRing &)

    cdef cppclass PBPolyFactory "DefaultRinged<PolynomialFactory> ":
        void operator=(const PBPolynomialFactory&)
        PBPoly operator()(long)
        PBPoly operator()(PBPoly)
        PBPoly operator()(PBMonom)

    cdef cppclass PBRefCounter:
        PBRefCounter()
        bint released()



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from random import Random
from sage.rings.polynomial.pbori.pbori import if_then_else as ite
from .PyPolyBoRi import Polynomial
from .statistics import used_vars_set


class CNFEncoder:
    def __init__(self, r, random_seed=16):
        self.random_generator = Random(random_seed)
        self.one_set = r.one().set()
        self.empty_set = r.zero().set()
        self.r = r

    def zero_blocks(self, f):
        r"""
        Divide the zero set of f into blocks.

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: r = declare_ring(["x", "y", "z"], dict())
            sage: from sage.rings.polynomial.pbori.cnf import CNFEncoder
            sage: e = CNFEncoder(r)
            sage: e.zero_blocks(r.variable(0)*r.variable(1)*r.variable(2))
            [{z: 0}, {y: 0}, {x: 0}]
        """
        f = Polynomial(f)
        variables = f.vars_as_monomial()

        space = variables.divisors()
        variables = list(variables.variables())
        zeros = f.zeros_in(space)
        rest = zeros
        res = []

        # inefficient compared to polynomials lex_lead
        def choose(s):
            indices = []
            assert not s.empty()
            nav = s.navigation()
            while not nav.constant():
                e = nav.else_branch()
                t = nav.then_branch()
                if e.constant() and not e.terminal_one():
                    indices.append(nav.value())
                    nav = t
                else:
                    if self.random_generator.randint(0, 1):
                        indices.append(nav.value())
                        nav = t

                    else:
                        nav = e
            assert nav.terminal_one()
            res = self.one_set
            for i in reversed(indices):
                res = ite(i, res, self.empty_set)
            return next(iter(res))
        while not rest.empty():
            l = choose(rest)
            l_variables = set(l.variables())

            def get_val(var):
                if var in l_variables:
                    return 1
                return 0
            block_dict = {v: get_val(v) for v in variables}

            l = l.set()
            self.random_generator.shuffle(variables)
            for v in variables:
                candidate = l.change(v.index())
                if candidate.diff(zeros).empty():
                    l = l.union(candidate)
                    del block_dict[v]
            rest = rest.diff(l)
            res.append(block_dict)
        return res

    def clauses(self, f):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: r = declare_ring(["x", "y", "z"], dict())
            sage: from sage.rings.polynomial.pbori.cnf import CNFEncoder
            sage: e = CNFEncoder(r)
            sage: sorted(e.clauses(r.variable(0)*r.variable(1)*r.variable(2)), key=lambda d: sorted(d.items()))
            [{z: 0, y: 0, x: 0}]
            sage: sorted(e.clauses(r.variable(Integer(1))+r.variable(Integer(0))), key=lambda d: sorted(d.items()))
            [{y: 0, x: 1}, {y: 1, x: 0}]
        """
        # we form an expression for a var configuration *not* lying in the
        # block it is evaluated to 0 by f, iff it is not lying in any zero
        # block of f+1
        return [{variable: 1 - value for variable, value in b.items()}
                for b in self.zero_blocks(f + 1)]

    def polynomial_clauses(self, f):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: r = declare_ring(["x", "y", "z"], dict())
            sage: from sage.rings.polynomial.pbori.cnf import CNFEncoder
            sage: e = CNFEncoder(r)
            sage: e.polynomial_clauses(r.variable(0)*r.variable(1)*r.variable(2))
            [x*y*z]
            sage: v = r.variable
            sage: p = v(1)*v(2)+v(2)*v(0)+1
            sage: groebner_basis([p], heuristic = False)==groebner_basis(e.polynomial_clauses(p), heuristic = False)
            True
        """

        def product(l):
            res = l[0]
            for p in l[1:]:
                res = res * p
            # please care about the order of these multiplications for
            # performance
            return res
        return [product([variable + value for (variable, value)
                         in b.items()]) for b in self.clauses(f)]

    def to_dimacs_index(self, v):
        return v.index() + 1

    def dimacs_encode_clause(self, c):
        def get_sign(value):
            if value == 1:
                return 1
            return -1

        items = sorted(c.items(), reverse=True)
        return " ".join([str(v) for v in [
            get_sign(value) * self.to_dimacs_index(variable)
            for (variable, value) in items] + [0]])

    def dimacs_encode_polynomial(self, p):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: d = {}
            sage: r = declare_ring(["x", "y", "z"], d)
            sage: from sage.rings.polynomial.pbori.cnf import CNFEncoder
            sage: e = CNFEncoder(r)
            sage: sorted(e.dimacs_encode_polynomial(d["x"]+d["y"]+d["z"]))
            ['-1 -2 -3 0', '-1 2 3 0', '1 -2 3 0', '1 2 -3 0']
        """
        return [self.dimacs_encode_clause(c) for c in self.clauses(p)]

    def dimacs_cnf(self, polynomial_system):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: r = declare_ring(["x", "y", "z"], dict())
            sage: from sage.rings.polynomial.pbori.cnf import CNFEncoder
            sage: e = CNFEncoder(r)
            sage: e.dimacs_cnf([r.variable(0)*r.variable(1)*r.variable(2)])
            'c cnf generated by PolyBoRi\np cnf 3 1\n-1 -2 -3 0'
            sage: e.dimacs_cnf([r.variable(1)+r.variable(0)])
            'c cnf generated by PolyBoRi\np cnf 3 2\n-1 2 0\n1 -2 0'
            sage: e.dimacs_cnf([r.variable(0)*r.variable(1)*r.variable(2), r.variable(1)+r.variable(0)])
            'c cnf generated by PolyBoRi\np cnf 3 3\n-1 -2 -3 0\n1 -2 0\n-1 2 0'
        """
        clauses_list = [c for p in polynomial_system for c in self.
                        dimacs_encode_polynomial(p)]
        res = ["c cnf generated by PolyBoRi"]
        r = polynomial_system[0].ring()
        n_variables = r.n_variables()
        res.append(f"p cnf {n_variables} {len(clauses_list)}")
        res.extend(clauses_list)
        return "\n".join(res)


class CryptoMiniSatEncoder(CNFEncoder):
    group_counter = 0

    def dimacs_encode_polynomial(self, p):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: d=dict()
            sage: r = declare_ring(["x", "y", "z"], d)
            sage: from sage.rings.polynomial.pbori.cnf import CryptoMiniSatEncoder
            sage: e = CryptoMiniSatEncoder(r)
            sage: p = d["x"]+d["y"]+d["z"]
            sage: p.deg()
            1
            sage: len(p)
            3
            sage: e.dimacs_encode_polynomial(p)
            ['x1 2 3 0\nc g 1 x + y + z']
            sage: e.dimacs_encode_polynomial(p+1)
            ['x1 2 -3 0\nc g 2 x + y + z + 1']
        """
        if p.deg() != 1 or len(p) <= 1:
            res = super().dimacs_encode_polynomial(p)
        else:
            invert_last = bool(p.has_constant_part())
            variables = list(p.vars_as_monomial().variables())
            indices = [self.to_dimacs_index(v) for v in variables]
            if invert_last:
                indices[-1] = -indices[-1]
            indices.append(0)
            res = ["x" + " ".join(str(v) for v in indices)]
        self.group_counter = self.group_counter + 1
        group_comment = f"\nc g {self.group_counter} {str(p)[:30]}"
        return [c + group_comment for c in res]

    def dimacs_cnf(self, polynomial_system):
        r"""

        TESTS::

            sage: from sage.rings.polynomial.pbori import *
            sage: r = declare_ring(["x", "y", "z"], dict())
            sage: from sage.rings.polynomial.pbori.cnf import CryptoMiniSatEncoder
            sage: e = CryptoMiniSatEncoder(r)
            sage: e.dimacs_cnf([r.variable(0)*r.variable(1)*r.variable(2)])
            'c cnf generated by PolyBoRi\np cnf 3 1\n-1 -2 -3 0\nc g 1 x*y*z\nc v 1 x\nc v 2 y\nc v 3 z'
            sage: e.dimacs_cnf([r.variable(1)+r.variable(0)])
            'c cnf generated by PolyBoRi\np cnf 3 1\nx1 2 0\nc g 2 x + y\nc v 1 x\nc v 2 y'
            sage: e.dimacs_cnf([r.variable(0)*r.variable(1)*r.variable(2), r.variable(1)+r.variable(0)])
            'c cnf generated by PolyBoRi\np cnf 3 2\n-1 -2 -3 0\nc g 3 x*y*z\nx1 2 0\nc g 4 x + y\nc v 1 x\nc v 2 y\nc v 3 z'
        """
        uv = list(used_vars_set(polynomial_system).variables())
        res = super().dimacs_cnf(polynomial_system)
        res += "\n" + "\n".join(f"c v {self.to_dimacs_index(v)} {v}"
                                for v in uv)
        return res



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
import gzip
from io import StringIO
import base64 as uu
import re
from types import ModuleType
from .PyPolyBoRi import Polynomial
AUTO = "auto"
SINGLE = "single"


# def ref_file_name(f):
#     name=re.sub("data/","",f)
#     name=sub(r"\.py","",name)
#     l=name.split("/")[:-1]


def reencode_blocks(block_str):
    return str(block_str).replace(",", "_")


def parse_blocks(block_str, data):
    if block_str == AUTO:
        return data.block_start_hints
    if block_str == SINGLE:
        return []
    return [int(i) for i in block_str.split(",")]


def load_ref_raw(s):
    s = re.sub("data/", "", s)
    s = re.sub(r"data\.", "", s)
    s = re.sub(r"\.py", "", s)
    s = re.sub(r"\.", "/", s)

    ref_file = "ref/" + s + ".ref"
    with open(ref_file) as res_f:
        res = res_f.read()
    return res


def load_ref(s, ordering='lp', blocks=SINGLE):
    return load_ref_gz_uu(s, ordering, blocks)


def ordering_suffix(o, blocks=None):
    if o == "lp":
        return ""
    if re.match("block", o):
        return "." + o + "_" + reencode_blocks(blocks)
    return "." + o


def number_of_declared_vars(data):
    try:
        return data.number_of_declared_vars
    except AttributeError:
        return data.r.ngens()


def load_ref_gz_uu(s, o, b):
    s = re.sub("data/", "", s)
    s = re.sub(r"data\.", "", s)
    s = re.sub(r"\.py", "", s)
    s = re.sub(r"\.", "/", s)

    ref_file = "ref/" + s + ordering_suffix(o, b) + ".ref.gz.uu"
    res = StringIO()
    uu.decode(ref_file, res)
    res = res.getvalue()
    res = StringIO(res)
    res = gzip.GzipFile(fileobj=res, mode='r').read()
    res = res.replace(" ", "")
    return res


def convert_refs(ref_file_orig):
    with open(ref_file_orig) as file:
        content = file.read()
    buf_out = StringIO()
    zipped = gzip.GzipFile(filename=ref_file_orig, mode='w', fileobj=buf_out)
    zipped.write(content)
    zipped.close()
    val = buf_out.getvalue()
    with open(ref_file_orig + ".gz.uu", "w") as out:
        uu.encode(out_file=out, in_file=StringIO(val))


def dyn_generate(content, name):
    module = ModuleType(name)
    import_header = """from .PyPolyBoRi import Variable,Monomial, Polynomial, Ring, OrderCode
from itertools import chain
from .blocks import AlternatingBlock,Block,AdderBlock,if_then,HigherOrderBlock,declare_ring as orig_declare_ring,declare_block_scheme,MacroBlock\n
def declare_ring(blocks, context=None):
  if context is None:
    context=globals()
  return orig_declare_ring(blocks,context)
"""
    exec(import_header + content, module.__dict__)
    if hasattr(module, "ideal"):
        module.ideal = [Polynomial(p) for p in module.ideal]
    return module


def clean_data(data):
    for a in dir(data):
        if a != "r":
            delattr(data, a)


def load_data(file_name, base_dir='./'):
    in_file = file_name
    if not re.match("^data", in_file):
        in_file = "data/" + in_file
    in_file = re.sub(r".py$", "", in_file)
    in_file = re.sub(r"\.", "/", in_file)
    in_file = in_file + ".py"
    with open(base_dir + in_file) as f:
        in_file = f.read()
    return dyn_generate(in_file, "pb_data")


def load_file(file_name):
    with open(file_name) as f:
        in_file = f.read()
    return dyn_generate(in_file, "pb_data")
